# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 Dirk Stöcker <trac@dstoecker.de>
# All rights reserved.
#
# This software is licensed as described in the file COPYING, which
# you should have received as part of this distribution. The terms
# are also available at https://trac.edgewall.com/license.html.
#
# This software consists of voluntary contributions made by many
# individuals. For the exact contribution history, see the revision
# history and logs, available at http://projects.edgewall.com/trac/.

import json
import urllib.request, urllib.parse, urllib.error
import urllib.request, urllib.error, urllib.parse

from trac.config import Option
from trac.core import Component, implements
from trac.util.html import tag
from trac.web.chrome import add_script

from tracspamfilter.api import _, user_agent
from tracspamfilter.captcha.api import ICaptchaMethod


class HcaptchaCaptcha(Component):
    """hCaptcha implementation"""

    implements(ICaptchaMethod)

    # use similar values as reCAPTCHA from Google
    private_key = Option('spam-filter', 'captcha_hcaptcha_private_key', '',
    """Sitekey for hCaptcha usage.""", doc_domain="tracspamfilter")

    public_key = Option('spam-filter', 'captcha_hcaptcha_public_key', '',
        """Secret for hCaptcha usage.""", doc_domain="tracspamfilter")

    def generate_captcha(self, req):
        add_script(req, 'https://js.hcaptcha.com/1/api.js')

        return None, tag(
            tag.div(**{
                'class': 'h-captcha',
                'data-sitekey': self.public_key
            }),
            tag.input(type='submit', value=_("Submit"))
        )

    def encode_if_necessary(self, s):
        if isinstance(s, str):
            return s.encode('utf-8')
        return s

    def verify_key(self, private_key, public_key):
        if private_key is None or public_key is None:
            return False
        # FIXME - Not yet implemented
        return True

    def verify_captcha(self, req):
        hcaptcha_response_field = req.args.get('h-captcha-response')
        remoteip = req.remote_addr
        try:
            params = urllib.parse.urlencode({
                'secret': self.encode_if_necessary(self.private_key),
                'remoteip': self.encode_if_necessary(remoteip),
                'response': self.encode_if_necessary(hcaptcha_response_field),
            }).encode("utf-8")
            request = urllib.request.Request(
                url='https://hcaptcha.com/siteverify',
                data=params,
                headers={
                    'Content-type': 'application/x-www-form-urlencoded',
                    'User-agent': user_agent
                }
            )

            response = urllib.request.urlopen(request)
            return_values = json.loads(response.read())
            response.close()
        except Exception as e:
            self.log.warning("Exception in hCaptcha handling (%s)", e)
        else:
            if return_values['success'] is True:
                return True
            else:
                self.log.warning("hCaptcha returned error: %s",
                                 return_values['error-codes'])
        return False

    def is_usable(self, req):
        return self.public_key and self.private_key
